/**   -=- DrawLines -=-
 *
 *    This program creates thumbnail sketches from coordinates of lines
 *    recorded in a log file generated during user experiments. From a
 *    graphical user interface, this program allows to load the log file,
 *    and to browse the lines viewing the result on an image. This thumbnail
 *    image can be saved in a file under the PNG format. An accumulation
 *    mode allows to add several lines on the same thumbnail image.
 *
 *    Copyright (C) 2006-2008 Guillaume RIVIERE
 *
 *    This file is part of DrawLines.
 *
 *    DrawLines is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 * 
 *    DrawLines is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 * 
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>
 **/


/**
 * Class DrawMap
 * - - - - - - -
 * Une zone de dessin qui herite de la classe Draw.
 * 
 * Inspire la discussion suivante :
 *    http://mail.gnome.org/archives/gtkmm-list/2005-October/msg00054.html
 *
 * Guillaume Riviere 30-06-2006
 */

#include <iostream>
#include <math.h>

#include "DrawLines.hh" 

using namespace std ;

/////////////////////////////////////////
//-- Constructeurs / Destructeur PUBLIC
/////////////////////////////////////////

DrawLines::DrawLines(int w, int h)
  : Draw (w, h),
    m_Xmin (0),
    m_Ymin (0),
    m_Xmax (0),
    m_Ymax (0),
    m_map_Xmin (0.0),
    m_map_Ymin (0.0),
    m_map_Xmax (0.0),
    m_map_Ymax (0.0),
    m_isDrawing (false),
    m_draw_image_border (true),
    m_line_expected_isDrawing (true)
{ 
  m_cut.ex = 0 ;
}

////////////////////////
//-- Accesseurs PUBLIC
////////////////////////

void DrawLines::set_map_coord (const double Xmin, const double Xmax,
			       const double Ymin, const double Ymax)
{
  m_map_Xmin = Xmin ;
  m_map_Xmax = Xmax ;
  m_map_Ymin = Ymin ;
  m_map_Ymax = Ymax ;

  queue_draw () ;
}

void DrawLines::set_picture_coord (const double Xmin, const double Xmax,
				   const double Ymin, const double Ymax)
{
  m_Xmin = Xmin ;
  m_Xmax = Xmax ;
  m_Ymin = Ymin ;
  m_Ymax = Ymax ;

  queue_draw () ;
}

//////////////////////
//-- Methodes PUBLIC
//////////////////////


void DrawLines::save_to_file (const string& filename)
{

  Glib::RefPtr<Gdk::Pixbuf> pixbuf ;
  Glib::RefPtr<Gdk::Window> win = get_window();  
  Glib::RefPtr<Gdk::GC> gc = get_style()->get_black_gc();  

  pixbuf = Gdk::Pixbuf::create ((const Glib::RefPtr<Gdk::Drawable>&)win, gc->get_colormap(), 0, 0, 0, 0, getWidth(), getHeight()) ;

  pixbuf->save (filename, "png") ;

}

void DrawLines::multiline_active (const bool active)
{
  m_multiline_active = active ;
}

void DrawLines::multiline_add (const double x1, const double y1,
			       const double x2, const double y2)
{
  struct cut c ;

  c.ex = 0 ; // Rien

  c.x1 = x1 ;
  c.y1 = y1 ;
  c.x2 = x2 ;
  c.y2 = y2 ;

  m_multiline_tab.push_back (c) ;
}

void DrawLines::multiline_clear ()
{
  m_multiline_tab.clear () ;
}

/////////////////////////
//-- Methodes PROTECTED
/////////////////////////

int DrawLines::x2graph (const double x)
{
  return (int) (getWidth() / (m_Xmax - m_Xmin) * (x - m_Xmin)) ;
}

int DrawLines::y2graph (const double y)
{
  return (int) (getHeight() / (m_Ymax - m_Ymin) * (y - m_Ymin)) ;
}

/**
 * Redessine la zone lors de l'appel d'vnement expose_event.
 */
bool DrawLines::on_expose_event(GdkEventExpose* event)  
{  
  double x1, y1, x2, y2 ;

  Draw::on_expose_event(event) ;

  Glib::RefPtr<Gdk::Window> win = get_window();  
  Glib::RefPtr<Gdk::GC> gc = get_style()->get_black_gc();  

  Gdk::Color black ("black") ;
  Gdk::Color blue ("blue") ;
  Gdk::Color red ("red") ;
  Gdk::Color green ("green") ;

  /* Dessiner le contour de l'image */
  if (m_draw_image_border) {
    gc->set_rgb_fg_color (black) ;
    win->draw_line(gc, 0, 0, 0, getHeight()-1);
    win->draw_line(gc, 0, getHeight()-1, getWidth()-1, getHeight()-1);
    win->draw_line(gc, getWidth()-1, getHeight()-1, getWidth()-1, 0);
    win->draw_line(gc, 0, 0, getWidth()-1, 0);
  }

  /* Dessiner le contour de la carte */
  gc->set_rgb_fg_color (green) ;
  win->draw_line(gc, x2graph(m_map_Xmin), y2graph(m_map_Ymin), x2graph(m_map_Xmin), y2graph(m_map_Ymax));
  win->draw_line(gc, x2graph(m_map_Xmin), y2graph(m_map_Ymax), x2graph(m_map_Xmax), y2graph(m_map_Ymax));
  win->draw_line(gc, x2graph(m_map_Xmax), y2graph(m_map_Ymax), x2graph(m_map_Xmax), y2graph(m_map_Ymin));
  win->draw_line(gc, x2graph(m_map_Xmin), y2graph(m_map_Ymin), x2graph(m_map_Xmax), y2graph(m_map_Ymin));

  /* Dessiner la ligne de coupe attendue */
  gc->set_rgb_fg_color (red) ;

  switch (m_cut.ex) {

  case 0 : // Exo2 : Rien
    break ;

  case 1 : // Exo1 no1 : X = 8
    x1 = 289 ;
    y1 = 39 ;
    x2 = 289 ;
    y2 = 355 ;
    win->draw_line(gc, x2graph(x1), y2graph(y1), x2graph(x2), y2graph(y2));
    break ;
 
  case 2 : // Exo1 no2 : Y = 2
    x1 = 66 ;
    y1 = 284 ;
    x2 = 374 ;
    y2 = 284 ;
    win->draw_line(gc, x2graph(x1), y2graph(y1), x2graph(x2), y2graph(y2));
    break ;
      
  case 3 : // Exo1 no3 : NE - SE
    x1 = 66 ;
    y1 = 355 ;
    x2 = 374 ;
    y2 = 45 ;
    win->draw_line(gc, x2graph(x1), y2graph(y1), x2graph(x2), y2graph(y2));
    break ;

  case 4 : // Exo1 no4 : // a no3 passant par inter no1 et no2 
    x1 = 218 ;
    y1 = 355 ;
    x2 = 374 ;
    y2 = 201 ;
    win->draw_line(gc, x2graph(x1), y2graph(y1), x2graph(x2), y2graph(y2));
    break ;

  case 5 : // Exo1 no5 : _|_ a no3 passant par inter no1 et no2 
    x1 = 66 ;
    y1 = 89 ;
    x2 = 370 ;
    y2 = 355 ;
    win->draw_line(gc, x2graph(x1), y2graph(y1), x2graph(x2), y2graph(y2));
    break ;

  case 6 : // Exo1 no6 : 1/2 S - inter no1 et no2
    x1 = 220 ;
    y1 = 355 ;
    x2 = 374 ;
    y2 = 197 ;
    win->draw_line(gc, x2graph(x1), y2graph(y1), x2graph(x2), y2graph(y2));    
    break ;

  default :
    cout << "Erreur type exercice" << endl ;    
  }

  /* Dessiner la ligne de coupe effectuee */
  if (m_isDrawing) {
    gc->set_rgb_fg_color (blue) ;

    /* Dessiner lignes cumulees */
    if (m_multiline_active) {
      for (vector<struct cut>::iterator it = m_multiline_tab.begin() ; it != m_multiline_tab.end() ; it++)
	win->draw_line(gc,x2graph(it->x1), y2graph(it->y1), x2graph(it->x2), y2graph(it->y2));
    }
    
    /* Dessiner ligne courante */
    win->draw_line(gc, x2graph(m_cut.x1), y2graph(m_cut.y1), x2graph(m_cut.x2), y2graph(m_cut.y2));
  }

  gc->set_rgb_fg_color (black) ;

  return true;  
}  


/**
 * Appel optimise de queue_draw_area() pour dessiner (redessiner) une droite.
 *
 * Redessine n sous-boites englobantes approcimant la droite passant par les points (xA, yA) et (xB, yB).
 *
 * Cas 1 : xB < xA et yB < yA
 * """"""""""""""""""""""""""
 * Descendre du point B au point A
 *
 *    + B            + B
 *     \             |_
 *      \              |_
 *       \               |_
 *        + A              |_+ A
 *
 *
 * Cas 2 : xB > xA et yB > yA
 * """"""""""""""""""""""""""
 * Descendre du point A au point B
 *
 *    + A            + A
 *     \             |_
 *      \              |_
 *       \               |_
 *        + B              |_+ B
 *
 * Cas 3 : xB < xA et yB < yA
 * """"""""""""""""""""""""""
 * Partir du point C pour remonter vers D
 *                         D_
 *        + B             _| + B
 *       /              _| 
 *      /            C_|
 *     /             |
 *    + A            + A
 *
 * C = (xA, yA+h)
 *
 * Cas 4 : xB < xA et yB < yA
 * """"""""""""""""""""""""""
 * Partir du point C pour remonter vers D
 *                         D_
 *        + A             _| + A
 *       /              _| 
 *      /            C_|
 *     /             |
 *    + B            + B
 *
 * C = (xB, yB+h)
 *
 */
void DrawLines::queue_draw_area_optimizedForLine (double xA, double yA, double xB, double yB, int n)
{
  double x, y, w, h ;
  double dx, dy ;

  /* Calcul des increments */
  if (xB < xA)
    {
      w = (xA - xB) / n ;
      if (yB < yA) { // cas 1 ( xB < xA et yB < yA)
	h = (yA - yB) / n ;
	dx = w ;
	dy = h ;
	x = xB ;
	y = yB ;
      }
      else { // cas 4 ( xB < xA et yB > yA)
	h = (yB - yA) / n ;
	dx = w ;
	dy = -h ;
	x = xB ;
	y = yB - h ;
      }
    }
  else
    {
      w = (xB - xA) / n ;
      if (yB < yA) { // cas 3 ( xB > xA et yB < yA)
	h = (yA - yB) / n ;
	dx = w ;
	dy = -h ;
	x = xA ;
	y = yA - h ;
      }
      else { // cas 2 ( xB > xA et yB > yA)
	h = (yB - yA) / n ;
	dx = w ;
	dy = h ;
	x = xA ;
	y = yA ;
      }
    }

  /* n appels de queue_draw_area() sur chacune des n sous-boites englobantes */
  for (int i=0 ; i<n ; i++) {
    this->queue_draw_area((int)round(x), (int)round(y), (int)round(w+1), (int)round(h+1)) ;
    x += dx ;
    y += dy ;
  }
}


/**
 * Calculer l'equation affine y=ax+b
 * passant par les points (x1,y1) et (x2, y2).
 */
bool DrawLines::equation_affine (double x1, double y1,
			       double x2, double y2,
			       double* a, double* b)
{
  if (x1 == x2)
    return false ;
  
  *a = (y1 - y2) / (x1 - x2) ;
  *b = y2 - (*a * x2) ;

  return true ;
}

/**
 * Calcul des points d'intersection (X1, Y1) et (X2, Y2)
 * de la droite d'equation Y = a * X + b
 * avec le cadre ancre en (0, 0) et en (w, h).
 *
 * RMQ : Coordonnees regardees a l'envers par rapport a la zone Gtk::DrawingArea
 * RMQ : (X1, Y1) et (X2, Y2) sont sur le cadre !
 *
 *    Y |
 *      |
 *      |
 *    h |_ _ _ _ _ _ _ _ _ _ _ _ (w,h)
 *      |                       |
 *      |                       |
 *      |                       |
 *      |                       |
 *      |_ _ _ _ _ _ _ _ _ _ _ _|_ _ _ _ _ _
 *    (0,0)                     w            X
 *
 */
void DrawLines::intersections_with_area (double bottom, double top,
				       double left, double right,
				       double a, double b,
				       int* x1, int* y1,
				       int* x2, int* y2)
{
  double X1, Y1, X2, Y2 ;

  if (a == 0) {
    X1 = left ;
    Y1 = b ;

    X2 = right ;
    Y2 = b ;
  }
  else {
    // Calcul du point (X1, Y1)
    X1 = right ;
    Y1 = a * X1 + b ;

    if (Y1 > top) { // Si la droite sort du cadre par le haut
      Y1 = top ;
      X1 = (Y1 - b) / a ;
    }
    else if (Y1 < bottom) { // Si la droite sort du cadre par le bas
      Y1 = bottom ;
      X1 = (Y1 - b) / a ;
    }

    // Calcul du point (X2, Y2)
    X2 = left ;
    Y2 = a * X2 + b ;
    
    if (Y2 > top) { // Si la droite sort du cadre par le haut
      Y2 = top ;
      X2 = (Y2 - b) / a ;
    }
    else if (Y2 < bottom) { // Si la droite sort du cadre par le bas
      Y2 = bottom ;
      X2 = (Y2 - b) / a ;
    }
  }
  
  *x1 = (int) round (X1) ;
  *y1 = (int) round (Y1) ;
  *x2 = (int) round (X2) ;
  *y2 = (int) round (Y2) ;
}
